#!/usr/bin/perl

# BSF: Bayesian Spam Filter
#  formerly known as ghlm: Gutierrez Herrera Lemus Martoq, the program
#  that pretends to be a bayesian spamfilter  :-)
#
#  Copyright GNU/GPL-2 2002,
#    by Cristian Gutierrez, Alvaro Herrera, Ricardo Lemus, Aldrin Martoq
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



###################################
######### CONFIGURE HERE ##########
###################################

# default dir where data is stored
$BSFDIR = $ENV{HOME}."/.bsf/";

# the basename for the files containing tokens and stats
# the names of the DB files are $BSFSTATFILE.dbinbox and $BSFSTATFILE.dbspam
$BSFSTATFILE = $BSFDIR."tokenstats";


# Cost of misclassifying an inocent mail as spam. Ex: misclassifying a mail
# as spam is as bad as receiving 9 SPAM into your inbox.
# Suggested: 9  (means Pbb(mail is spam)> 90% )
# See http://www.ics.forth.gr/~potamias/mlnia/paper_2.pdf
# FIXME: Translate this to an adaptative cost?
$SPAMCOST = 30;

# max tokens selected to calculate bayes rule (see runtest subroutine)
$MAXTOKENS = 80;

# max length of word for a token
$MAXTOKENLENGTH = 10;

# if you want (some) messages displayed when running interactive.
# Posible values: 0=none, 1=general messages, 2=internal debugging
# **NOTE**: if STDIN is used, no messages are generated.
$BSFDEBUG = 1;





use DB_File;



##############################
######### SUBROUTINES ########
##############################



############
# Show usage
#
# args: none
# returns: nothing

sub usage {
    print STDERR <<EOF
Usage:
	$0 <command> [files..]

<command> must be one of:
	addinbox	Add tokens from file(s) to inbox statistics
	addspam		Add tokens from file(s) to spam  statistics
	reminbox	Remove inbox statistics from file(s)
	remspam		Remove spam  statistics from file(s)
	inbox2spam	Move statistics from inbox to spam
	spam2inbox	Move statistics from spam  to inbox

	test		Run the test on the files specified,
			returns 0 (inbox) or 1 (spam)
	testadd		Same as above, but add the statistics to the
			appropiate folder.

	If no files are specified, standard input is used and no messages are
	displayed.
EOF
;
}

##################
# String tokenizer
#   Extracts tokens from the readen contents of a file.
#
# args:
#   $inputref      reference to the array of lines from the file
# returns:
#   \%tokens       reference to a token hash; keys as tokens, times as values.

sub tokenizer {
    my $inputref = shift;
    my %tokens;

    # DEACTIVATED  1word tokenizer
    if (0) {
	
	$tokens{$_}++ for map lc, map /([a-zA-Z\'0-9\-]+)/g, @{$inputref};

	return \%tokens;
    }
   
    # DEACTIVATED  2 words tokenizer
    if (0) {
	my $tok;

	my $prev;

	foreach my $t (map lc, map /([a-zA-Z0-9]+)/g, @{$inputref}) {
	    $t = sprintf ("%.".$MAXTOKENLENGTH."s", $t);

	    if (defined $prev) {
		$tok = $prev.$t;
		$tokens{$tok}++;
	    }
	    $prev = $t." ";
	}
    }

    # DEACTIVATED  3 words tokenizer
    if (0) {
	my $tok;
	my $p0, $p1;
	foreach my $l (@{$inputref}) {

	    foreach my $t (map lc, map /([a-zA-Z0-9]+)/g, $l) {
		$t = sprintf ("%.".$MAXTOKENLENGTH."s", $t);
		
		if (defined $p0 && defined $p1) {
		    $tok = $p0." ".$p1." ".$t;
		    $tokens{$tok}++;
		}
		$p0 = $p1;
		$p1 = $t;
	    }
	}
    }

    # ACTIVATED   3 words hash tokenizer (long)
    if (1) {
	my $tok;
	my $h, $c;
	foreach my $l (@{$inputref}) {
	    my $p0, $p1;
	    my $P0, $P1, $T;

	    foreach my $t (map lc, map /([a-zA-Z0-9]+)/g, $l) {

		$T = $t if ($BSFDEBUG > 1);
		$h = pack("L", 0);
		while ($t ne "") {
		    $c = chop ($t);
		    $h <<= 2;
		    $h += ord ($c);
		}
		if (defined $p0 && defined $p1) {
		    $tok = pack("SCC", sort ($p0, $p1, $h));
		    #$tok = pack("LSS", $p0, $p1, $h);
		    #$tok = pack("LLL", $p0, $p1, $h);
		    
		    $tokens{$tok}++;

		    if ($BSFDEBUG > 1) {
			my $s1, $c1, $c2;
			($s1, $c1, $c2) = unpack("SCC", $tok);
			printf STDERR "%5d %3d %3d => $P0 $P1 $T\n", $s1, $c1, $c2;
		    }
		}
		$p0 = $p1;
		$p1 = $h;

		if ($BSFDEBUG > 1) {
		    $P0 = $P1;
		    $P1 = $T;
		}

	    }
	}
    }


    return \%tokens;
}



#################
# Read file
#   Read contents from a file, or STDIN. Returns lines from the file in an
#   array.
#
# args:
#   $file           string with filename.
# returns:
#   \@contents      reference to the array of lines.

sub readfile {
    my $file = shift;
    my @contents = ();
    my $state = 0;

    if (! $file || $file eq "-") {
	<STDIN>;
	while (<STDIN>) {
	    next if ($_ =~ /^\s+/ && $state==1);
	    next if ($_ =~ /^Date\:/);
	    last if ($_ =~ /^$/);
	    $state = 0;
	    push @contents, $_;
	    if ($_ =~ /^Received\:/) {
		$state = 1;
	    }
	}
	while (<STDIN>) { 
	    push @contents, $_;
	}
    } else {
	open (A, $file) or die "failed to open file $file: $!";
	<A>;
	while (<A>) {
	    next if ($_ =~ /^\s+/ && $state==1);
	    next if ($_ =~ /^Date\:/);
	    last if ($_ =~ /^$/);
	    push @contents, $_;
	    $state = 0;
	    if ($_ =~ /^Received\:/) {
		$state = 1;
	    }
	}
	while (<A>) { 
	    push @contents, $_;
	}
	close (A);
    }
    return \@contents;
}



##########################
# Load statistics
#   Load the entire file into memory. The hash values for the tokenstats are
#   arrays: First element is cached probability, Second element is times seen
#   in inbox, Third element is times seen in spam.
#
# args: none
# returns:
#   $ninbox        number of inbox in database
#   $nspam         number of spam  in database
#   \@tokenstats   reference to an array,
#                  first element has inbox tokens, second has spam tokens.

sub loadstats {
    my ($ninbox, $nspam);
    my @tokenstats;
    my ($p, $g, $b, $t);
    my %tokinbox;
    my %tokspam;

    if (! -d $BSFDIR) {
	mkdir $BSFDIR, 0700 or die "cannot create directory $BSFDIR: $!";
    }
    my $dbt = new DB_File::BTREEINFO;
    #$dbt->{'cachesize'} = 2 * 1024 * 1024;
    my $di;
    my $ds;
    $di = tie (%tokinbox, "DB_File", $BSFSTATFILE.".dbinbox", O_RDWR|O_CREAT, 0600, $dbt) or die "dbinbox : $!\n";
    $ds = tie (%tokspam, "DB_File", $BSFSTATFILE.".dbspam", O_RDWR|O_CREAT, 0600, $dbt) or die "dbspam : $!\n";

    $di->filter_fetch_value ( sub { $_ = unpack("S", $_) } );
    $di->filter_store_value ( sub { $_ = pack("S", $_) } );

    $ds->filter_fetch_value ( sub { $_ = unpack("S", $_) } );
    $ds->filter_store_value ( sub { $_ = pack("S", $_) } );

    open (A, "$BSFSTATFILE") or goto loadstats_failed;
    $ninbox = <A> + 0;
    $nspam = <A> + 0;

    goto loadstats_ok;

loadstats_failed:
    $ninbox = 0; $nspam = 0;

loadstats_ok:
    close (A);

    push @tokenstats, \%tokinbox;
    push @tokenstats, \%tokspam;

    return ($ninbox, $nspam, \@tokenstats);
}



#################
# Save statistics
#   Save (updated) statistics. If a token has zero values, it's eliminated.
#   If $numinbox, $numspam <=0; no tokens are saved.
#
# args:
#   $numinbox        number of inbox
#   $numspam         number of spam
#   $ts              reference to the tokenstats array.
# returns: nothing

sub savestats {
    my $numinbox = shift;
    my $numspam = shift;
    my $ts = shift;

    open (A, "> $BSFSTATFILE") or die "cannot create file $BSFSTATFILE: $!";
    print A $numinbox."\n";
    print A $numspam."\n";

    print STDERR "= ". (keys %{$$ts[0]}). ", ". (keys %{$$ts[1]}). " tokens saved\n" if ($BSFDEBUG);

    if (1) {
	untie %{$$ts[0]};
	untie %{$$ts[1]};
    }
    close (A);
}


############################
# Run Command
#   Run the specified command, update statistics, etc etc.
#
# FIXME: What *should* happen if a token does not exists and it's tried to be
#        removed ???
# FIXME: What *should* happen if $ninbox, $nspam < 0 ???
#
# args:
#   $cmd            String. One of: addinbox, addspam, reminbox, remspam,
#                   inbox2spam, spam2inbox.
#   $ninbox         number of inbox
#   $nspam          number of spam
#   $ftref          reference to the file's tokens
#   $tsref          reference to the token statistics array
# returns:
#   $ninbox         updated number of inbox
#   $nspam          updated number of spam
#   $found          0 if no spam found, 1 if found
#   $needsavestats  0 if no stats changed, 1 if so

sub runcommand {
    my $cmd = shift;                  # command
    my $ninbox = shift;               # numinbox
    my $nspam = shift;                # numspam
    my $ftref = shift;                # ref to filetokens
    my $tsref = shift;                # ref to tokenstats array
    my $found = 0;                    # 1 if spam was found ($val == 1)
    my $needsavestats = 0;            # 0 if no stats changed, 1 if so

    my $nftkeys = keys %{$ftref};
    printf STDERR "%6d tokens  " , $nftkeys if ($BSFDEBUG);

    my $toki = $$tsref[0];
    my $toks = $$tsref[1];

    # omit if no tokens found in the file
    return ($ninbox, $nspam, $found, $needsavestats) if ($nftkeys == 0);

    if ($cmd eq "addinbox") {
	$ninbox++;
	foreach my $k (keys %{$ftref}) {
	    $$toki{$k} = 0 if (! defined ($$toki{$k}));
	    $$toki{$k}++;
	}
	$needsavestats = 1;
    } elsif ($cmd eq "reminbox") {
	$ninbox--;
	foreach my $k (keys %{$ftref}) {
	    next if (! defined ($$toki{$k}));
	    $$toki{$k}--;
	    delete $$toki{$k} if ($$toki{$k} <= 0);
	}
	$needsavestats = 1;
    } elsif ($cmd eq "addspam") {
	$nspam++;
	foreach my $k (keys %{$ftref}) {
	    $$toks{$k} = 0 if (! defined ($$toks{$k}));
	    $$toks{$k}++;
	}
	$needsavestats = 1;
    } elsif ($cmd eq "remspam") {
	$nspam--;
	foreach my $k (keys %{$ftref}) {
	    next if (! defined ($$toks{$k}));
	    $$toks{$k}--;
	    delete $$toks{$k} if ($$toks{$k} <= 0);
	}
	$needsavestats = 1;
    } elsif ($cmd eq "inbox2spam") {
	$ninbox--;
	$nspam++;
	foreach my $k (keys %{$ftref}) {

	    $$toks{$k} = 0 if (! defined ($$toks{$k}));
	    $$toks{$k}++;

	    next if (! defined ($$toki{$k}));
	    $$toki{$k}--;
	    delete $$toki{$k} if ($$toki{$k} <= 0);
	}
	$needsavestats = 1;
    } elsif ($cmd eq "spam2inbox") {
	$ninbox++;
	$nspam--;
	foreach my $k (keys %{$ftref}) {

	    $$toki{$k} = 0 if (! defined ($$toki{$k}));
	    $$toki{$k}++;

	    next if (! defined ($$toks{$k}));
	    $$toks{$k}--;
	    delete $$toks{$k} if ($$toks{$k} <= 0);

	}
	$needsavestats = 1;
    } elsif ($cmd eq "test") {
	my ($val, $score) = getbayesscore ($ninbox, $nspam, $ftref, $tsref);
	$found = $val;

	print STDERR "$val" if ($BSFDEBUG);
    } elsif ($cmd eq "testadd") {
	my ($val, $score) = getbayesscore ($ninbox, $nspam, $ftref, $tsref);
	$found = $val;
	my $ncmd;
	if ($val == 0) {
	    $ncmd = "addinbox";
	} else {
	    $ncmd = "addspam";
	}
	print STDERR "$ncmd\t$val" if ($BSFDEBUG);
	my $oldBSFDEBUG = $BSFDEBUG;
	$BSFDEBUG = 0;
	($ninbox, $nspam) = runcommand ($ncmd, $ninbox, $nspam, $ftref, $tsref);
	$BSFDEBUG = $oldBSFDEBUG;
	$needsavestats = 1;
    } elsif ($cmd eq "show") {
	my %toklist;
	foreach my $k (keys %{$toki}) {
	    $toklist{$k} = [0, 0, $k] if (! defined ($toklist{$k}) );
	    $toklist{$k}[0] += $$toki{$k};
	}

	foreach my $k (keys %{$toks}) {
	    $toklist{$k} = [0, 0, $k] if (! defined ($toklist{$k}) );
	    $toklist{$k}[1] += $$toks{$k};
	}

	foreach my $k (keys %toklist) {
	    $p = $toklist{$k}[2];
	    
	    print $toklist{$k}[0]."\t".$toklist{$k}[1];
	    foreach (unpack ("SSS", $p)) {
		print "\t".$_;
	    }
	    print "\n";
	}
    } else {
	if ($BSFDEBUG) {
	    print STDERR "Unknown command: $cmd\n";
	    usage ();
	}
	exit 0;
    }
	

    $ninbox = 0 if ($ninbox < 0);
    $nspam = 0 if ($nspam < 0);

    print STDERR "\n" if ($BSFDEBUG);
    return ($ninbox, $nspam, $found, $needsavestats);
}


sub max { my ($a, $b) = @_; $a > $b ? $a : $b; }
sub min { my ($a, $b) = @_; $a < $b ? $a : $b; }



############################
# Test the file if it belongs to the spam class or not
#    
# args:
#   $ninbox         number of inbox
#   $nspam          number of spam
#   $ftref          reference to the file tokens
#   $tsref          reference to the token statistics array
# returns:
#   $val            0 (inbox) or 1 (spam)
#   $score          score (0-100)

sub getbayesscore {
    my $ninbox = shift;
    my $nspam = shift;
    my $ftref = shift;
    my $tsref = shift;



    if ($ninbox == 0 or $nspam == 0) {
	print STDERR "$ninbox $nspam ZERO!\n" if ($BSFDEBUG>1);
	return (0, 0);
    }
    my %delta;
    my %tokprob;

    my $g, $b, $pbad, $pgood, $ptot, $val_aux;

    my $toki = $$tsref[0];
    my $toks = $$tsref[1];

    foreach my $tok (keys %{$ftref}) {

	# FIXME: > 5 deberia ser por cuantos mail tenemos...
	#if (defined ($tsref->{$tok}) && ($tsref->{$tok}[0] + $tsref->{$tok}[1]) > 5) {

	$g = 0;
	$b = 0;
	$g = $$toki{$tok} * 2 if (defined ($$toki{$tok}) );
	$b = $$toks{$tok} + 0 if (defined ($$toks{$tok}) );


	# only add known tokens
	if (($g + $b) > 0) {
	    
	    $pbad = min (1.0, (1.0 * $b / $nspam) );
	    $pgood = min (1.0, ($g / $ninbox) );
	    $ptot = $pbad + $pgood;
	    $val_aux = min (0.99, 1.0 * $pbad/$ptot);
	    $tokprob{$tok} = max (0.01, $val_aux);

	    $delta{$tok} = abs (1.0 * $tokprob{$tok} - 0.5);
	} else {
#	    $tokprob{$tok} = 0.01;
#	    $delta{$tok} = abs (1.0 * $tokprob{$tok} - 0.5);
	}
    }

    # choose the $MAXTOKENS tokens which are farer from 0.5
    my @sorted = sort {$delta{$b} <=> $delta{$a} } keys %delta;
    my @selected;
    
    foreach my $tok (@sorted) {
	last if ($delta{$tok} < 0.3);
	push @selected, $tok;
    }

    if ($#selected < $MAXTOKENS) {
	@selected = @sorted;
    }
    
    printf STDERR "%5d used ", $#selected if ($BSFDEBUG);
    print STDERR "\n\tUsing $#selected tokens:" if ($BSFDEBUG > 1);

    my $prod = 1.0;
    my $prod_pond = 1.0;

    foreach my $tok (@selected) {
	$prod *= $tokprob{$tok};
	$prod_pond *= (1 - $tokprob{$tok});
    }

    my $val;

    if ($prod + $prod_pond == 0) {
	$val = 50;
    } else {
	$val = ($prod / ($prod + $prod_pond))*100;
    }

    print STDERR map { my ($s1, $c1, $c2) = unpack("SCC", $_); sprintf "\n%5d %3d %3d => $tokprob{$_}", $s1, $c1, $c2; } @selected if ($BSFDEBUG > 1);
    print STDERR "\n\tBAYES = ".$val."\n" if ($BSFDEBUG > 1);
    
    return ( $val > 100*$SPAMCOST/(1+$SPAMCOST) ? 1 : 0, $val);
}


###############################################
########### MAIN PROGRAM STARTS HERE ##########
###############################################


# Check args and get command
if ( $#ARGV<0 ) {
    usage ();
    exit 0;
}
my $command = shift @ARGV;

# Load statistics
my ($numinbox, $numspam, $tokenstats) = loadstats ();

# How many mails has been processed
# How many spam has been found
# How many tokens has been read
# How many updates has been made
my ($numprocessed, $numspamfound, $numtokens, $needsavestats) = (0, 0, 0, 0);

# if STDIN, use "-" as filename and disable messages
if ($#ARGV < 0) {
    push @ARGV, "-";
    $BSFDEBUG = 0;
}

# For each file specified, run command
foreach my $file (@ARGV) {
    print STDERR uc($command)."\t\"".$file."\"\t" if ($BSFDEBUG);
    print STDERR "\n" if ($BSFDEBUG>1);

    my $filetokens = tokenizer (readfile ($file));
    my $ntoken = keys %{$filetokens};

    my $nfound, $nneedsave;
    ($numinbox, $numspam, $nfound, $nneedsave) = runcommand ($command, $numinbox, $numspam, $filetokens, $tokenstats);

    $numspamfound += $nfound;
    $needsavestats += $nneedsave;
    $numtokens += $ntoken;

    $numprocessed++;
}

print STDERR "Total tokens READ     = $numtokens\n" if ($BSFDEBUG);
print STDERR "Num inbox             = $numinbox\n" if ($BSFDEBUG);
print STDERR "Num spam              = $numspam\n" if ($BSFDEBUG);
print STDERR "Mails processed       = $numprocessed\n" if ($BSFDEBUG);
print STDERR "Spam found            = $numspamfound\n" if ($BSFDEBUG);
print STDERR "Saving stats ...      " if ($BSFDEBUG);

savestats ($numinbox, $numspam, $tokenstats);
undef $tokenstats;

my $exitcode = 0;
if ($numspamfound == $numprocessed && $numspamfound > 0) {
    $exitcode = 1;
}

print STDERR "Exit code is          = $exitcode\n" if ($BSFDEBUG);
exit ($exitcode);
